#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include "basic.h"
#include "poly.h"

/* Fork a subtask, return pointer to thread */

static count_t count = 0 ;
static pthread_mutex_t *mutex ;

void *run_stub(void *z) {
  count_t r = run_subtask(z) ;
  free_arg(z) ;
  lock_mutex(mutex) ;
  count += r ;
  unlock_mutex(mutex) ;
  return NULL ;
}

pthread_t *fork_subtask(subtask_t *z) {
  pthread_t *th = malloc_check(sizeof(*th)) ;
  create_thread(th,run_stub,z) ;
  return th ;
}

/* List of threads */

typedef struct cell {
  pthread_t *fst ;
  struct cell *next ;
} cell_t ;

void free_cell(cell_t *p) { return free(p) ; }

cell_t *alloc_cell(pthread_t *fst, cell_t *next) {
  cell_t *r = malloc(sizeof(*r)) ;
  if (!r) {
    perror("malloc") ;
    exit(2) ;
  }
  r->fst = fst ;
  r->next = next ;
  return r ;
}


/* Fork a new thread to run subtask */
static cell_t *threads = NULL ;

void emit_thread(subtask_t *z) {
  z = copy_arg(z) ;
  threads = alloc_cell(fork_subtask(z),threads) ;  
}

count_t zyva(int n, int depth) {
  mutex = alloc_mutex() ;
  count = 0 ;
    
  /* Fork all subtasks */
  int ntasks = generate_subtasks(n,depth,emit_thread) ;
  if (verbose) fprintf(stderr,"Subtasks: %i\n",ntasks) ;

  /* Sum results */
  for (cell_t *p = threads ; p ; p = p->next) {
    (void)join_thread(p->fst) ;
  }
  return count ;
}

void usage(char *pgm)
{
  fprintf(stderr,"usage: %s [-v]*\n",pgm);
  exit(-1);
}

int main(int argc, char **argv) {
  char *arg=0;
  char *pgm = argv[0];
  int n = 20 ;
  int gen_now = 15 ;

  while (--argc) {
    arg = *++argv;
    if (*arg == '-') {
      arg++ ;
       if (*arg == 'g') {
        if (arg[1]) {
          gen_now += atoi(arg+1) ;
        } else {
          usage(pgm) ;
        }
       } else if (*arg == 'v') {
         verbose++;
       } else {
         usage(pgm);
       }
    } else break ;      
  }
  if (argc >= 1) n = atoi(arg);
  if (n > NMAX) {
      fprintf(stderr,"argument must be <= %d\n",NMAX);
      exit(2);
  }
  count_t r = zyva(n,gen_now) ;
  printf("%" PRC "\n",r) ;
  return 0 ;
}
